IMPLEMENTATION MODULE Calendar;
(* $Z- *)
FROM Storage  IMPORT MainHeap, HeapAllocate, HeapAvail,
                     HeapDeallocate, Available, ALLOCATE, DEALLOCATE;
FROM Lib      IMPORT WordFill;
FROM SYSTEM   IMPORT SIZE, Seg;
FROM Str      IMPORT Length;
FROM IO       IMPORT WrLn, WrStr;
TYPE
  tDate     = POINTER TO RECORD
                Year, Month, Day, Guard : CARDINAL;
              END;
  tDaysInMonth = ARRAY [0..12] OF CARDINAL;
CONST
  C1900 = 1900;
  daysInMonths = tDaysInMonth(0,31,59,90,120,151,181,212,243,273,304,334,365);
VAR totalHeap : CARDINAL;
    adrHeap   : ADDRESS;
  (* --- internal procedures ----------------------------------------------*)
  PROCEDURE checkDateGuard (proc : ARRAY OF CHAR;
                            date : tDate);
  BEGIN
    IF NOT (date^.Guard = Seg(date^)) THEN
      WrLn; WrStr ('Error in procedure '); WrStr (proc);
      WrStr (': Variable of type tDate not yet created');
      HALT;
    END;
  END checkDateGuard;
  (*-----------------------------------------------------------------------*)
  PROCEDURE allocDate (procName : ARRAY OF CHAR;
                              VAR date     : tDate);
  BEGIN
    (* IF NOT (date^.Guard = Seg(date^)) THEN *)
      IF NOT Available(SIZE(date)) THEN
        WrLn; WrStr ('Not enough Memory for procedure: ');
        WrStr (procName);
        HALT;
      ELSE
        NEW(date); date^.Guard := Seg(date^);
      END;
    (* END; *)
  END allocDate;
  (*-----------------------------------------------------------------------*)
  PROCEDURE leapYear (year : CARDINAL) : BOOLEAN;
  BEGIN
    RETURN ((year MOD 4) = 0) AND
           ((year MOD 100) # 0) OR ((year MOD 400) = 0);
  END leapYear;
  (*-----------------------------------------------------------------------*)
  PROCEDURE totalDaysInYear (year : CARDINAL) : CARDINAL;
  BEGIN
    RETURN 365 + ORD(leapYear(year));
  END totalDaysInYear;
  (*-----------------------------------------------------------------------*)
  PROCEDURE nrDayInYear (year, month, day : CARDINAL) : CARDINAL;
  BEGIN
    RETURN  daysInMonths[month-1] + day +
            ORD((month > 2) AND (leapYear(year)));
   END nrDayInYear;
(*-------------------------------------------------------------------------*)
  PROCEDURE totalDays (year, month, day : CARDINAL) : LONGCARD;
  VAR
    y1 : LONGCARD;
  BEGIN
    y1 := LONGCARD(year - 1);
    RETURN  (  y1 * 365 + y1 DIV 4 - y1 DIV 100 + y1 DIV 400
             + LONGCARD(nrDayInYear(year, month, day)) - 1);
  END totalDays;
(*-------------------------------------------------------------------------*)
  PROCEDURE weekd (year, month, day : CARDINAL) : eWeekday;
  BEGIN
    RETURN eWeekday(CARDINAL(totalDays(year, month, day) MOD 7));
  END weekd;
(*-------------------------------------------------------------------------*)
  PROCEDURE daysInMonth (    year, month : CARDINAL) : CARDINAL;
  BEGIN
    IF (month = 2) AND leapYear (year) THEN
      RETURN 29;
    ELSE
      RETURN daysInMonths[month]-daysInMonths[month-1];
    END;
  END daysInMonth;
(*-------------------------------------------------------------------------*)
  PROCEDURE validDate (year, month, day : CARDINAL) : BOOLEAN;
  VAR
    dateOK : BOOLEAN;
  BEGIN
    dateOK :=            (month >= 1) AND (month <= 12);
    dateOK := dateOK AND (year >= 1) AND (year <= 9999);
    dateOK := dateOK AND (day >= 1) AND (day <= daysInMonth(year, month));
    RETURN dateOK;
  END validDate;
(* --- exported procedures ------------------------------------------------*)
  PROCEDURE DeallocDate (VAR date : tDate);
  BEGIN
    IF (date^.Guard = Seg(date^)) THEN
      date^.Guard := 0;
      DISPOSE(date);
    END;
  END DeallocDate;
  (*-----------------------------------------------------------------------*)
  PROCEDURE CopyDate (dateIn : tDate; VAR dateOut : tDate);
  BEGIN
    checkDateGuard ('CopyDate', dateIn);
    allocDate ('CopyDate', dateOut);
    dateOut^ := dateIn^;
  END CopyDate;
  (*-----------------------------------------------------------------------*)
  PROCEDURE CompDate (date1, date2 : tDate) : eRelation;
  BEGIN
    checkDateGuard ('CompDate', date1);
    checkDateGuard ('CompDate', date2);
    IF (date1^.Year # date2^.Year) THEN
      RETURN VAL (eRelation, ORD(date1^.Year > date2^.Year));
    END;
    IF (date1^.Month # date2^.Month) THEN
      RETURN VAL (eRelation, ORD(date1^.Month > date2^.Month));
    END;
    IF (date1^.Day # date2^.Day) THEN
      RETURN VAL (eRelation, ORD(date1^.Day > date2^.Day));
    END;
    RETURN equal;
  END CompDate;
  (*-----------------------------------------------------------------------*)
  PROCEDURE DateDiff (date1, date2 : tDate) : LONGINT;
  BEGIN
    checkDateGuard ('DateDiff', date1);
    checkDateGuard ('DateDiff', date2);
    RETURN LONGINT(totalDays(date1^.Year, date1^.Month, date1^.Day))
         - LONGINT(totalDays(date2^.Year, date2^.Month, date2^.Day));
  END DateDiff;
  (*-----------------------------------------------------------------------*)
  PROCEDURE StrToDate (    string   : ARRAY OF CHAR;
                           modeYear : eDateMode;
                       VAR date     : tDate         ) : BOOLEAN;
  VAR
    i, nrOfCharInYear, length,
    day, month, year   : CARDINAL;
    dateOK             : BOOLEAN;
    dateStatus         : (start, days, firstStop, months, secondStop, years);
   (* ---------------------------------- *)
   PROCEDURE charToDigit (ch : CHAR) : CARDINAL;
     BEGIN
       RETURN ORD(ch) - ORD('0');
     END charToDigit;
   (* ---------------------------------- *)
  BEGIN
    length := Length(string);
    IF length < 5 THEN
      RETURN FALSE;
    END;
    FOR i := 0 TO length - 1 DO
      IF (string[i] # '.') AND
         ((string[i] < '0') OR (string[i] > '9')) THEN
        RETURN FALSE;
      END;
    END;
    dateOK := TRUE; nrOfCharInYear := 0; dateStatus := start;
    FOR i := 0 TO length - 1 DO
      CASE dateStatus OF
        | start :
          IF string[i] = '.' THEN
            RETURN FALSE;
          END;
          day := charToDigit(string[i]);
          dateStatus := days;
        | days :
          IF (string[i] = '.') THEN
            dateStatus := firstStop;
          ELSE
            day := day * 10 + charToDigit(string[i]);
          END;
        | firstStop :
          IF string[i] = '.' THEN
            RETURN FALSE;
          END;
          month := charToDigit(string[i]);
          dateStatus := months;
        | months    :
          IF (string[i] = '.') THEN
            dateStatus := secondStop;
          ELSE
            month := month * 10 + charToDigit(string[i]);
          END;
        | secondStop:
          IF string[i] = '.' THEN
            RETURN FALSE;
          END;
          year := charToDigit(string[i]);
          dateStatus := years;
          INC(nrOfCharInYear);
        | years     :
          IF (string[i] = '.') OR (year > 999) THEN
            RETURN FALSE;
          END;
          year := year * 10 + charToDigit(string[i]);
          INC(nrOfCharInYear);
      END;
    END;
    IF dateStatus # years THEN
      RETURN FALSE;
    END;
    IF (modeYear = short) AND (nrOfCharInYear < 3) THEN
      INC(year, C1900);
    END;
    IF validDate (year, month, day) THEN
      allocDate ('StrToDate', date);
      date^.Day   := day;
      date^.Month := month;
      date^.Year  := year;
      RETURN TRUE;
    ELSE
      RETURN FALSE;
    END;
  END StrToDate;
(*-------------------------------------------------------------------------*)
  PROCEDURE CardToDate (    year, month, day : CARDINAL;
                            modeYear : eDateMode;
                        VAR date     : tDate              ) : BOOLEAN;
  TYPE
  BEGIN
    IF (modeYear = short) AND (year <= 99) AND (year > 1) THEN
      INC(year, C1900);
    END;
    IF  validDate(year, month, day) THEN
      allocDate ('CardToDate', date);
      date^.Day := day;
      date^.Month := month;
      date^.Year := year;
      RETURN TRUE;
    ELSE
      RETURN FALSE;
    END;
  END CardToDate;
(*-------------------------------------------------------------------------*)
   PROCEDURE DateToStr (    date : tDate;
                            modeDayMonth : eDateMode;
                            modeYear     : eDateMode;
                        VAR string : ARRAY OF CHAR);
   VAR
     i, year : CARDINAL;
   (* ---------------------------------- *)
   PROCEDURE digitToChar (d : CARDINAL) : CHAR;
     BEGIN
       RETURN CHAR(d + ORD('0'));
     END digitToChar;
   (* ---------------------------------- *)
   BEGIN
     checkDateGuard ('DateToStr', date);
     i := 0;
     IF (date^.Day >= 10) OR (modeDayMonth = long) THEN
       string[0] := digitToChar (date^.Day DIV 10); INC(i);
     END;
     string[i] := digitToChar (date^.Day MOD 10); INC(i);
     string[i] := '.'; INC(i);
     IF (date^.Month >= 10) OR (modeDayMonth = long) THEN
       string[i] := digitToChar (date^.Month DIV 10); INC(i);
     END;
     string[i] := digitToChar (date^.Month MOD 10); INC(i);
     string[i] := '.'; INC(i);
     year := date^.Year;
     IF (year <= C1900) OR (year > 1999) THEN
       modeYear := long;
     END;
     IF (modeYear = long) THEN
       string [i]:= digitToChar (year DIV 1000); INC(i);
       year := year MOD 1000;
       string[i] := digitToChar (year DIV 100); INC(i);
     END;
     year := year MOD 100;
     string[i] := digitToChar (year DIV 10); INC(i);
     string [i] := digitToChar (year MOD 10);
     IF (i < HIGH(string)) THEN
       INC(i); string[i] := 0C;
     END;
   END DateToStr;
(*-------------------------------------------------------------------------*)
   PROCEDURE DateToCard (    date : tDate;
                         VAR year, month, day : CARDINAL);
   BEGIN
     checkDateGuard ('DateToCard', date);
     year := date^.Year;
     month := date^.Month;
     day := date^.Day;
   END DateToCard;
(*-------------------------------------------------------------------------*)
  PROCEDURE DaysInMonth (    year, month : CARDINAL;
                         VAR days : CARDINAL) : BOOLEAN;
  BEGIN
    IF NOT validDate (year, month, 1) THEN
      RETURN FALSE;
    ELSE
      days := daysInMonth (year, month);
      RETURN TRUE;
    END;
  END DaysInMonth;
(*-------------------------------------------------------------------------*)
   PROCEDURE DateToDayInYear (    date : tDate;
                              VAR day  : CARDINAL);
   BEGIN
     checkDateGuard ('DayInYear', date);
     day := nrDayInYear (date^.Year, date^.Month, date^.Day);
   END DateToDayInYear;
  (*-----------------------------------------------------------------------*)
  PROCEDURE DayInYearToDate (    year, day : CARDINAL;
                                 modeYear  : eDateMode;
                             VAR date      : tDate) : BOOLEAN;
  VAR
    i, leapAdd : CARDINAL;
  BEGIN
    IF (year <= 99) AND (modeYear = short) THEN
      INC(year, C1900);
    END;
    IF (day < 1) OR (day > totalDaysInYear(year)) THEN
      RETURN FALSE;
    END;
    IF (year < 1) OR (year > 9999) THEN
      RETURN FALSE;
    END;
    leapAdd := ORD(leapYear(year) AND (day >= 60));
    i := 13;
    REPEAT
      DEC(i);
    UNTIL (day > daysInMonths[i] + leapAdd);
    allocDate ('DayInYearToDate', date);
    date^.Year := year;
    date^.Month := i + 1;
    IF day = 60 THEN
      leapAdd := 0;
    END;
    date^.Day := day - daysInMonths[i] - leapAdd;
    RETURN TRUE;
  END DayInYearToDate;
(*-------------------------------------------------------------------------*)
  PROCEDURE DateAdd (    dateIn  : tDate;
                         days    : LONGINT;
                     VAR dateOut : tDate) : BOOLEAN;
  VAR
    allDays, daysInOutYear, outYear : LONGINT;
  BEGIN
    checkDateGuard ('DateAdd', dateIn);
    allDays := totalDays (dateIn^.Year, dateIn^.Month, dateIn^.Day);
    allDays := allDays + days + 1;
    IF (allDays < 1) THEN
      RETURN FALSE;
    END;
    outYear := allDays * 400 DIV 146097 + 1;
    (* 146097 = 400*365 + 100 - 4 + 1 wegen:
       totalDays = year*365 + year DIV 4 - year DIV 100 + year DIV 400 ===>
       totalDays = year*365*400 + 100*year DIV 400 -
                   4*year DIV 400 + 1*year DIV 400 *)
    daysInOutYear := allDays - LONGINT(totalDays(CARDINAL(outYear), 1, 1));
    RETURN DayInYearToDate (CARDINAL(outYear),CARDINAL(daysInOutYear),
                                                       long, dateOut);
  END DateAdd;
  (*-----------------------------------------------------------------------*)
  PROCEDURE Weekday (    date : tDate;
                     VAR weekday : eWeekday);
  BEGIN
    checkDateGuard ('WeekDay', date);
    weekday := weekd (date^.Year, date^.Month, date^.Day);
  END Weekday;
(*-------------------------------------------------------------------------*)
  PROCEDURE Week (    date : tDate;
                  VAR week, year : CARDINAL);
  VAR weekday0101 : eWeekday;
  BEGIN
    checkDateGuard ('Week', date);
    year := date^.Year;
    weekday0101 := weekd(year, 1, 1);
    DateToDayInYear(date, week);
    week := (week + ORD(weekday0101)-1) DIV 7 + ORD(weekday0101 <= thursday);
    IF (week = 0) THEN
      DEC(year);
      week := 52 + ORD(weekd(year, 1, 1) <= thursday);
    ELSIF (week = 53) AND (weekd (year+1, 1, 1) <= thursday) THEN
        week := 1; INC(year);
    END;
  END Week;
(*-------------------------------------------------------------------------*)
  PROCEDURE WeekBeginEnd (    week, year  : CARDINAL;
                              modeYear    : eDateMode;
                          VAR dateBeg, dateEnd : tDate) : BOOLEAN;
  VAR
    weekday0101     : eWeekday;
    days, totalDays : CARDINAL;
    bool            : BOOLEAN;
  BEGIN
    IF (year <= 99) AND (modeYear = short) THEN
      INC(year, C1900);
    END;
    weekday0101 := weekd (year, 1, 1);
    IF (week < 1) OR (week > 53) OR
       ((week = 53) AND (weekday0101 > thursday)) THEN
      RETURN FALSE;
    END;
    days := 7*week-ORD(weekday0101);
    IF (weekday0101 > thursday) THEN
      INC(days, 7);
    END;
    totalDays := totalDaysInYear(year);
    IF (days > totalDays) THEN
      bool := DayInYearToDate(year, days-6, modeYear, dateBeg);
      INC(year);
      bool := DayInYearToDate(year, days-totalDays, modeYear, dateEnd);
    ELSIF (days < 7) THEN
      bool := DayInYearToDate(year, days, modeYear, dateEnd);
      DEC(year);
      totalDays := totalDaysInYear(year);
      bool := DayInYearToDate(year, days+totalDays-6, modeYear, dateBeg);
    ELSE
      bool := DayInYearToDate(year, days, modeYear, dateEnd);
      bool := DayInYearToDate(year, days-6, modeYear, dateBeg);
    END;
    RETURN TRUE;
  END WeekBeginEnd;
(*-------------------------------------------------------------------------*)
  PROCEDURE MovableCeremon (    movableCeremon : eMovableCeremon;
                                year           : CARDINAL;
                                modeYear       : eDateMode;
                            VAR date           : tDate);
  VAR
    weekday : eWeekday;
    ylp1, ylp2, r1, r2, daysToAshWdnday : CARDINAL;
    bool                             : BOOLEAN;
  BEGIN
    IF (modeYear = short) AND (year <=99) THEN
      INC (year, C1900);
    END;
    IF movableCeremon = repentanceDay THEN
      weekday := weekd (year, 11, 16);
      bool := CardToDate (year, 11, 25 - ORD(weekday), long, date);
      IF (date^.Day > 22) THEN
        DEC(date^.Day, 7);
      END;
      RETURN;
    END;
    IF (movableCeremon = firstAdvent) OR
       (movableCeremon = secondAdvent)   THEN
      bool := CardToDate (year, 12, 24, long, date);
      DateToDayInYear (date, r1);
      weekday := weekd (year, 12, 24);
      IF movableCeremon = firstAdvent THEN
        DEC(r1, 22);
      ELSE
        DEC(r1,15);
      END;
      bool := DayInYearToDate (year, r1 - ORD(weekday), long, date);
      RETURN;
    END;
    (* AshWednesday (Easter-46) as basis *)
    ylp1 := (year DIV 100) - (year DIV 400) + 4;
    ylp2 := ylp1  + 11 - (year DIV 300);
    r1 := ((year MOD 19) * 19 + ylp2) MOD 30;
    r2 := (((year MOD 4) * 2 + (4 * year) +
            (6 * r1) + ylp1) MOD 7) + r1 + 22;
    IF  (r2=57) OR
       ((r1=28) AND (r2=56) AND ((11*(ylp2+1) MOD 30) < 19)) THEN
       DEC (r2, 7);
    END;
    IF leapYear(year) THEN
      INC (r2, 14);
    ELSE
      INC (r2, 13);
    END;
    CASE movableCeremon OF
      | ashWednesday : daysToAshWdnday := 0;
      | mothersDay :
        weekday := weekd (year, 5, 1);
        IF (nrDayInYear(year, 5, 14-ORD(weekday)) = r2+95) THEN
          daysToAshWdnday := 88;
        ELSE
          bool := CardToDate (year, 5, 14-ORD(weekday), long, date);
          RETURN;
        END;
      | goodFriday       : daysToAshWdnday :=  44;
      | easterSunday     : daysToAshWdnday :=  46;
      | easterMonday     : daysToAshWdnday :=  47;
      | ascensionDay     : daysToAshWdnday :=  85;
      | whitsunday       : daysToAshWdnday :=  95;
      | whitsunMonday    : daysToAshWdnday :=  96;
      | corpusChristiDay : daysToAshWdnday := 106;
      ELSE
    END;
    bool := DayInYearToDate (year, r2+daysToAshWdnday, long, date);
  END MovableCeremon;
END Calendar.
