IMPLEMENTATION MODULE Mouse;
  FROM SYSTEM IMPORT Registers;
  FROM Lib    IMPORT Intr, FatalError;
  FROM Graph  IMPORT HercWidth, HercDepth, Width, Depth;

  VAR mouseActive, graph : BOOLEAN;
      regs               : Registers;
      i                  : CARDINAL;
      savePatch          : BYTE;
(*--- internal -------------------------------------------------------------*)
   PROCEDURE checkActive;
   BEGIN
     IF NOT mouseActive THEN
       FatalError ('TSMouse not Initialised by Call to Reset !');
     END;
   END checkActive;
(*--- exported -------------------------------------------------------------*)
  PROCEDURE PatchHerc ();
  VAR adr : RECORD CASE : CARDINAL OF
                     | 0 : o, s : CARDINAL;
                     | 1 : p    : POINTER TO BYTE;
            END;   END;
  BEGIN
    (* Patch for Hercules *)
    IF (Width = HercWidth) AND (Depth = HercDepth) THEN
      adr.s := 40H;
      adr.o := 49H;
      IF adr.p^ # BYTE(6) THEN
        savePatch := adr.p^;
        adr.p^ := BYTE(6);
        regs.AX := 0;
        Intr (regs, 33H);
      END;
    END;
  END PatchHerc;
(*--------------------------------------------------------------------------*)
  PROCEDURE ResetPatchHerc ();
  VAR adr : RECORD CASE : CARDINAL OF
                     | 0 : o, s : CARDINAL;
                     | 1 : p    : POINTER TO CHAR;
            END;   END;
  BEGIN
    (* Patch for Hercules *)
    IF (Width = HercWidth) AND (Depth = HercDepth) THEN
      adr.s := 40H;
      adr.o := 49H;
      adr.p^ := savePatch;
    END;
  END ResetPatchHerc;
(*--------------------------------------------------------------------------*)
  PROCEDURE Reset ( VAR buttons : CARDINAL) : BOOLEAN;
  BEGIN
    regs.AX := 0;
    Intr (regs, 33H);
      IF INTEGER(regs.AX) = 0 THEN
        RETURN FALSE;
      ELSE
        mouseActive := TRUE;
        buttons := regs.BX;
        RETURN TRUE;
      END;
  END Reset;
(*--------------------------------------------------------------------------*)
  PROCEDURE CursorOn;
  BEGIN
    checkActive;
    regs.AX := 1;
    Intr (regs, 33H);
  END CursorOn;
(*--------------------------------------------------------------------------*)
  PROCEDURE CursorOff;
  BEGIN
    checkActive;
    regs.AX := 2;
    Intr (regs, 33H);
  END CursorOff;
(*--------------------------------------------------------------------------*)
  PROCEDURE Read ( VAR pressedButtons : sButtons;
                        VAR x, y           : CARDINAL);
  VAR but : RECORD CASE : BOOLEAN OF
                     | TRUE  : c : CARDINAL;
                     | FALSE : b : sButtons;
            END;   END;
  BEGIN
    checkActive;
    regs.AX := 3;
    Intr (regs, 33H);
    but.c := regs.BX;
    pressedButtons := but.b;
    x := regs.CX;
    y := regs.DX;
    IF NOT graph THEN
      x := x DIV 8;
      y := y DIV 8;
    END;
  END Read;
(*--------------------------------------------------------------------------*)
  PROCEDURE SetCursor ( x, y : CARDINAL);
  BEGIN
    checkActive;
    regs.AX := 4;
    regs.CX := x;
    regs.DX := y;
    IF NOT graph THEN
      regs.CX := regs.CX * 8;
      regs.DX := regs.DX * 8;
    END;
    Intr (regs, 33H);
  END SetCursor;
(*--------------------------------------------------------------------------*)
  PROCEDURE InformButtonPressed (    button : eButtons;
                                 VAR buttonState : eButtonState;
                                 VAR nrPressed : CARDINAL;
                                 VAR x, y : CARDINAL );
  BEGIN
    checkActive;
    regs.AX := 5;
    regs.BX := ORD (button);
    Intr (regs, 33H);
    buttonState := eButtonState (ORD(button) IN BITSET(regs.AX));
    nrPressed := regs.BX;
    x := regs.CX;
    y := regs.DX;
    IF NOT graph THEN
      x := x DIV 8;
      y := y DIV 8;
    END;
  END InformButtonPressed;
(*--------------------------------------------------------------------------*)
  PROCEDURE InformButtonReleased (    button : eButtons;
                                  VAR buttonState : eButtonState;
                                  VAR nrPressed : CARDINAL;
                                  VAR x, y : CARDINAL );
  BEGIN
    checkActive;
    regs.AX := 6;
    regs.BX := ORD (button);
    Intr (regs, 33H);
    buttonState := eButtonState (ORD(button) IN BITSET(regs.AX));
    nrPressed := regs.BX;
    x := regs.CX;
    y := regs.DX;
    IF NOT graph THEN
      x := x DIV 8;
      y := y DIV 8;
    END;
  END InformButtonReleased;
(*--------------------------------------------------------------------------*)
  PROCEDURE SetXScope       ( xLeft, xRight : CARDINAL );
  BEGIN
    checkActive;
    regs.AX := 7;
    regs.CX := xLeft;
    regs.DX := xRight;
    IF NOT graph THEN
      regs.CX := regs.CX * 8;
      regs.DX := regs.DX * 8;
    END;
    Intr (regs, 33H);
  END SetXScope;
(*--------------------------------------------------------------------------*)
  PROCEDURE SetYScope       ( yTop, yBottom : CARDINAL );
  BEGIN
    checkActive;
    regs.AX := 8;
    regs.CX := yTop;
    regs.DX := yBottom;
    IF NOT graph THEN
      regs.CX := regs.CX * 8;
      regs.DX := regs.DX * 8;
    END;
    Intr (regs, 33H);
  END SetYScope;
(*--------------------------------------------------------------------------*)
  PROCEDURE DefineGraphCursor    (     graphCursor : rGraphCursor );
  VAR adr : RECORD CASE : CARDINAL OF
                     | 0 : o, s : CARDINAL;
                     | 1 : a    : ADDRESS;
                     | 2 : p    : POINTER TO BYTE;
            END;   END;
  BEGIN
    regs.AX := 9;
    adr.a := ADR (graphCursor.masks);
    regs.ES := adr.s;
    regs.DX := adr.o;
    regs.BX := graphCursor.xHot;
    regs.CX := graphCursor.yHot;
    Intr (regs, 33H);
    graph := TRUE;
  END DefineGraphCursor;
(*--------------------------------------------------------------------------*)
  PROCEDURE DefineTextCursor     ( type : eCursorType;
                                   screenMask, cursorMask : CARDINAL);
  BEGIN
    checkActive;
    regs.AX := 10;
    regs.BX := ORD(type);
    regs.CX := screenMask;
    regs.DX := cursorMask;
    Intr (regs, 33H);
    graph := FALSE;
  END DefineTextCursor;
(*--------------------------------------------------------------------------*)
  PROCEDURE GetMickeys           ( VAR dx, dy : INTEGER);
  BEGIN
    checkActive;
    regs.AX := 11;
    Intr (regs, 33H);
    dx := regs.CX;
    dy := regs.DX;
  END GetMickeys;
(*--------------------------------------------------------------------------*)
  PROCEDURE SetInterrupt    ( mouseCondi : sCondi;
                                   interProc  : ADDRESS );
  VAR int : RECORD CASE : BOOLEAN OF
                     | TRUE  : c : CARDINAL;
                     | FALSE : m : sCondi;
            END;   END;
      adr : RECORD CASE : BOOLEAN OF
                     | TRUE  : a    : ADDRESS;
                     | FALSE : o, s : CARDINAL;
            END;   END;
  BEGIN
    checkActive;
    regs.AX := 12;
    int.m := mouseCondi;
    regs.CX := int.c;
    adr.a := interProc;
    regs.ES := adr.s;
    regs.DX := adr.o;
    Intr (regs, 33H);
  END SetInterrupt;
(*--------------------------------------------------------------------------*)
  PROCEDURE SetLightpenOn        ();
  BEGIN;
    checkActive;
    regs.AX := 13;
    Intr (regs, 33H);
  END SetLightpenOn;
(*--------------------------------------------------------------------------*)
  PROCEDURE SetLightpenOff       ();
  BEGIN;
    checkActive;
    regs.AX := 14;
    Intr (regs, 33H);
  END SetLightpenOff;
(*--------------------------------------------------------------------------*)
  PROCEDURE SetMickeysPerPixel   ( dx, dy : INTEGER);
  BEGIN
    checkActive;
    regs.AX := 15;
    regs.CX := dx;
    regs.DX := dy;
    Intr (regs, 33H);
  END SetMickeysPerPixel;
(*--------------------------------------------------------------------------*)
  PROCEDURE SetInvisScope   ( xLeft, xRight, yTop, yBottom : CARDINAL);
  BEGIN
    checkActive;
    regs.AX := 16;
    regs.CX := xLeft;
    regs.DX := yTop;
    regs.SI := xRight;
    regs.DI := yBottom;
    Intr (regs, 33H);
  END SetInvisScope;
(*--------------------------------------------------------------------------*)
  PROCEDURE SetThreshold         ( speed : CARDINAL );
  BEGIN
    checkActive;
    regs.AX := 19;
    regs.DX := speed;
    Intr (regs, 33H);
  END SetThreshold;
(*--------------------------------------------------------------------------*)
BEGIN
  mouseActive := FALSE;
  graph := FALSE;
  WITH graphCursor DO
    xHot := 3;
    yHot := 0;
    WITH masks DO
      FOR i := 0 TO 15 DO
        screen [i] := {0..15};
      END;
      cursor [ 0] := {12};
      cursor [ 1] := {11, 13};
      cursor [ 2] := {11, 13};
      cursor [ 3] := {11, 13};
      cursor [ 4] := {11, 13};
      cursor [ 5] := {7..9, 11, 13};
      cursor [ 6] := {5..7, 9..11, 13..15};
      cursor [ 7] := {3..5, 7, 9, 11, 13, 15};
      cursor [ 8] := {3, 5, 7, 9, 11, 13, 15};
      cursor [ 9] := {3, 13, 15};
      cursor [10] := {3, 13, 15};
      cursor [11] := {3, 13, 15};
      cursor [12] := {3, 15};
      cursor [13] := {4, 14};
      cursor [14] := {5, 14};
      cursor [15] := {5, 14};
    END;
  END;
END Mouse.
