IMPLEMENTATION MODULE XGraph;
IMPORT IO, Graph;
  FROM Graph   IMPORT Width, Depth, PlotProc, PointProc, HLineProc,
                      Plot, Point, HLine, Line,
                      InitHerc, InitCGA, InitEGA, InitVGA;
  FROM FIO     IMPORT Exists, File, Open, Close, AssignBuffer, RdBin;
  FROM Storage IMPORT ALLOCATE;
  FROM SYSTEM  IMPORT In, Registers;
  FROM MATHLIB IMPORT Sqrt;
  FROM Lib     IMPORT Intr;
  FROM Str     IMPORT Length;

  TYPE SHORTBITSET = SET OF [SHORTCARD(0)..SHORTCARD(7)];
       BITARRAY    = ARRAY [0..7] OF SHORTBITSET;
       VECTOR      = ARRAY [1..95] OF CARDINAL;
       VECTORSHT   = ARRAY [1..3000] OF SHORTCARD;

  VAR adr, adrAscii, adrExtAscii : RECORD CASE : CARDINAL OF
                                     | 0 : A        : POINTER TO CARDINAL;
                                     | 1 : P        : POINTER TO BITARRAY;
                                     | 2 : Off, Seg : CARDINAL;
                                   END; END;
      extAscii : BOOLEAN;
      oldPlot  : PlotProc;
      oldPoint : PointProc;
      oldHLine : HLineProc;

      normalGraph, clipIndi : BOOLEAN;
      dxl, dyl, dxr, dyr : CARDINAL;

      chHeight, chExpan, chSpace : LONGREAL;
      xChUp, yChUp, xChWid, yChWid : LONGREAL;
      xSpace, ySpace : LONGREAL;
      chH, chE : CARDINAL;
      needAdjust : BOOLEAN;

(*- internal procedures ----------------------------------------------------*)

PROCEDURE myPlot (x, y : CARDINAL; color : CARDINAL);
BEGIN
  IF (x > Width-1) THEN RETURN; END;
  IF (y > Depth-1) THEN RETURN; END;
  IF clipIndi THEN
    IF (x < dxl) OR (x > dxr) THEN RETURN; END;
    IF (y < dyl) OR (y > dyr) THEN RETURN; END;
  END;
  y := Depth - y - 1;
  oldPlot (x, y, color);
END myPlot;

(*--------------------------------------------------------------------------*)

PROCEDURE myPoint (x, y : CARDINAL) : CARDINAL;
BEGIN
  RETURN oldPoint (x, Depth - y - 1);
END myPoint;

(*--------------------------------------------------------------------------*)

PROCEDURE myHLine (x1, y, x2 : CARDINAL; fillColor : CARDINAL);
BEGIN
  IF (y > Depth) THEN RETURN; END;
  IF (x1 > Width) THEN x1 := Width; END;
  IF (x2 > Width) THEN x2 := Width; END;
  IF clipIndi THEN
    IF (y < dyl) OR (y > dyr) THEN RETURN; END;
    IF (x1 < dxl) THEN x1 := dxl; END;
    IF (x2 > dxr) THEN x2 := dxr; END;
  END;
  IF (x1 <= x2) THEN
    y := Depth - y - 1;
    oldHLine (x1, y, x2, fillColor);
  END;
END myHLine;

(*--------------------------------------------------------------------------*)

PROCEDURE getPixels (ordCh : CARDINAL) : BITARRAY;
BEGIN
  IF (ordCh > 127) AND NOT extAscii THEN
    ordCh := ORD('*');
  END;
  IF (ordCh <= 127) THEN
    adr.Seg := adrAscii.Seg;
    adr.Off := adrAscii.Off + ordCh * 8;
  ELSE
    adr.Seg := adrExtAscii.Seg;
    adr.Off := adrExtAscii.Off + (ordCh-128) * 8;
  END;
  RETURN adr.P^;
END getPixels;

(*--------------------------------------------------------------------------*)

PROCEDURE adjustChar();
  VAR help, helpx, helpy : LONGREAL;
BEGIN
  chH := TRUNC(chHeight + 0.5);
  IF chH < 1 THEN chH := 1; END;
  chE := TRUNC(chExpan * chHeight + 0.5);
  IF chE < 1 THEN chE := 1; END;
  help := chHeight / Sqrt(xChUp*xChUp + yChUp*yChUp);
  helpx := xChUp * help;
  helpy := yChUp * help;
  xChWid :=  helpy * chExpan;
  yChWid :=  helpx;
  xChUp  := -helpx * chExpan;
  yChUp  :=  helpy;
  xSpace := xChWid * 8.0 * (chSpace + 1.0);
  ySpace := xChUp  * 8.0 * (chSpace + 1.0);
  needAdjust := FALSE;
END adjustChar;

(*--------------------------------------------------------------------------*)

PROCEDURE plotChar(x, y : INTEGER; ordCh : CARDINAL; color : CARDINAL);
VAR pixels : BITARRAY;
    i, j, k, l, xa, x0, y0 : CARDINAL;
BEGIN
  IF (x < 0) OR (y < 0) THEN
    RETURN;
  ELSE
    x0 := CARDINAL(x) + chE * 7;
    y0 := CARDINAL(y) + chH * 7;
  END;
  xa := x0;
  pixels := getPixels (ordCh);
  FOR k := 1 TO chH DO
    FOR l := 1 TO chE DO
      FOR j := 0 TO 7 DO
        FOR i := 0 TO 7 DO
          IF (SHORTCARD(i) IN pixels[j]) THEN
            IF normalGraph THEN
              Plot (x0 - i * chE, y0 - j * chH, color);
            ELSE
              Plot (x0 - i * chE, y0 + j * chH, color);
            END;
          END;
        END;
      END;
      INC(x0);
    END;
    x0 := xa;
    INC(y0);
  END;
END plotChar;

(*- exported procedures ----------------------------------------------------*)

PROCEDURE AutoInit() : tCardType;
VAR
  i, j : CARDINAL;
  first : SHORTCARD;
  reg : Registers;
  sevenIn : BOOLEAN;
  addr :
   RECORD CASE : CARDINAL OF
           | 1 : Off, Seg : CARDINAL;
           | 2 : p        : POINTER TO SHORTCARD;
         END;   END;
BEGIN
  normalGraph:=FALSE;
  (* Is it Hercules ? *)
  Intr(reg, 17);
  IF (4 IN BITSET(reg.AX)) AND (5 IN BITSET(reg.AX)) THEN
    first := In(3BAH);
    sevenIn := 7 IN BITSET(first);
    i := 0;
    j := 0;
    REPEAT
      first := In(3BAH);
      IF (7 IN BITSET(first)) # sevenIn THEN
        INC(i);
        IF i = 10 THEN
          InitHerc();
          RETURN Hercules;
        END;
      END;
      INC(j);
    UNTIL (j = 10000);
    InitCGA; (* Default *)
    RETURN notFound;
  END;
  (* Is it CGA ? *)
  addr.Off := 00A8H; addr.Seg := 0040H;
  IF addr.p^ = 0 THEN
    InitCGA();
    RETURN CGA;
  END;
  (* Is it EGA or VGA ? *)
  reg.AX := 1A00H;
  reg.BX := 0FEDCH;
  Intr(reg, 16);
  IF reg.BX = 0FEDCH THEN
    InitEGA();
    RETURN EGA;
  ELSE
    InitVGA();
    RETURN VGA;
  END;
END AutoInit;

(*--------------------------------------------------------------------------*)

PROCEDURE InitNormalGraph;
BEGIN
  IF NOT normalGraph THEN
    oldPlot  := Plot;
    oldPoint := Point;
    oldHLine := HLine;
    Plot     := myPlot;
    Point    := myPoint;
    HLine    := myHLine;
    normalGraph := TRUE;
  END;
END InitNormalGraph;

(*--------------------------------------------------------------------------*)

PROCEDURE ClearScreen();
VAR i : CARDINAL;
BEGIN
  FOR i := 0 TO Depth-1 DO
    HLine (0, i, Width-1, 0);
  END;
END ClearScreen;

(*--------------------------------------------------------------------------*)

PROCEDURE SetClipIndi (clInd : BOOLEAN);
BEGIN
  clipIndi := clInd;
END SetClipIndi;

(*--------------------------------------------------------------------------*)

PROCEDURE SetDevView (xld, yld, xru, yru : CARDINAL);
BEGIN
  dxl := xld; dyl := yld;
  dxr := xru; dyr := yru;
END SetDevView;

(*--------------------------------------------------------------------------*)

PROCEDURE PlotStr (x, y : CARDINAL; str : ARRAY OF CHAR; color : CARDINAL);
VAR pixels         : BITARRAY;
    k, ordCh       : CARDINAL;
    x0, y0, dx, dy : INTEGER;
    width          : LONGREAL;
BEGIN
  x0 := VAL(INTEGER, x);
  y0 := VAL(INTEGER, y);
  IF needAdjust THEN
    adjustChar();
  END;
  dx := INTEGER(TRUNC(xSpace));
  dy := INTEGER(TRUNC(ySpace));
  FOR k := 0 TO Length(str)-1 DO
    plotChar(x0, y0, ORD(str[k]), color);
    INC (x0, dx);
    INC (y0, dy);
  END;
END PlotStr;

(*--------------------------------------------------------------------------*)

PROCEDURE SetCharHeight (CharHeight : REAL);
BEGIN
  chHeight := VAL(LONGREAL, CharHeight);
  needAdjust := TRUE;
END SetCharHeight;

(*--------------------------------------------------------------------------*)

PROCEDURE SetCharExpan (CharExpan : REAL );
BEGIN
  chExpan := VAL(LONGREAL, CharExpan);
  needAdjust := TRUE;
END SetCharExpan;

(*--------------------------------------------------------------------------*)

PROCEDURE SetCharSpace (CharSpace : REAL);
BEGIN
  chSpace := VAL(LONGREAL, CharSpace);
  needAdjust := TRUE;
END SetCharSpace;

(*--------------------------------------------------------------------------*)

PROCEDURE SetCharUp (xCharUp, yCharUp : REAL);
BEGIN
  xChUp := VAL(LONGREAL,xCharUp);
  yChUp := VAL(LONGREAL,yCharUp);
  needAdjust := TRUE;
END SetCharUp;

(*--------------------------------------------------------------------------*)

BEGIN
  adrAscii.Seg := 0F000H; adrAscii.Off := 0FA6EH;
  adr.Seg := 0000H; adr.Off := 007EH;
  adrExtAscii.Seg := adr.A^;
                    adr.Off := 007CH;
  adrExtAscii.Off := adr.A^;
  IF (adrExtAscii.Off = 0) AND (adrExtAscii.Seg = 0) THEN
    extAscii := FALSE;
  ELSE
    extAscii := TRUE;
  END;
  normalGraph := FALSE;
  SetClipIndi (FALSE);
  SetDevView (0, 0, MAX(CARDINAL), MAX(CARDINAL));
  SetCharHeight (1.0);
  SetCharExpan (1.0);
  SetCharSpace (0.0);
  SetCharUp (0.0, 1.0);
END XGraph.
